# -*- coding: utf-8 -*-

"""
Doga.config.configer

This module gives other modules access to use configuration values
"""

import os
from ConfigParser import SafeConfigParser

# configuration file path
config_file_path = os.path.join(os.path.dirname(__file__), 'config.ini')
config_file = os.path.abspath(config_file_path)

# read config data using ConfigParser module
parser = SafeConfigParser()
parser.read(config_file)


def value(key):
    """ return config value for key

    param: key(str) : config data key
    return: config value for the requested key(str)
    """

    return parser.get('Doga', key)



# -*- coding: utf-8 -*-

"""
Doga.interfaces.gui

This module manages Doga GUI and keep it updated in realtime.
"""

import npyscreen


class WindowForm(npyscreen.ActionForm):

    def create(self, *args, **keywords):
        super(WindowForm, self).create(*args, **keywords)
    
    def while_waiting(self):
        pass
        
        
class DogaGUI(npyscreen.NPSApp):

    def __init__(self, statistics):
        # npyscreen form widget, parent of all widgets in GUI
        self.window = None

        # different data values to show in GUI
        self.alert_status = None
        self.doga_status = None
        self.doga_logs = None
        self.alert_history = None

        # reference to Statistics module functions that return data values
        self.template = statistics.template
        self.value = statistics.value

    def while_waiting(self):
        """ defines the default actions to be performed while waiting for user
        here this function updates widget data values periodically
        """

        # updating data values for all section
        self.alert_status.value = "%s" % self.template('alert')
        self.doga_status.value = "%s" % self.template('stats')
        self.doga_logs.values = self.value('logs')
        self.alert_history.values = self.value('history')

        # reverse lists to show them like recently updated(latest first)
        self.doga_logs.values = self.doga_logs.values[::-1]
        self.alert_history.values = self.alert_history.values[::-1]

        # display updated values in widgets, call 'Widget.display()' method
        self.alert_status.display()
        self.doga_status.display()
        self.doga_logs.display()
        self.alert_history.display()
        
    def main(self):
        """ 'main' function of npyscreen.NPSApp application object
        this function help in initial widget setup and rendering
        """

        # time(ms) to wait for user interactions
        self.keypress_timeout_default = 10

        # Form widget instance
        self.window = WindowForm(parentApp=self, name="Doga : HTTP Log Monitor",)
        
        # setup a section for Alert status
        self.alert_status = self.window.add(npyscreen.TitleText, name="Alert Status", max_height=3)
        self.alert_status.value = ""
        self.alert_status.editable = False

        # setup a section for Doga status
        self.doga_status = self.window.add(npyscreen.TitleText, name="Doga Status", max_height=3, rely=4)
        self.doga_status.value = ""
        self.doga_status.editable = False

        # setup a section for Doga logs
        self.doga_logs = self.window.add(npyscreen.BoxTitle, name="Doga Logs", max_width=50, relx=2, rely=7)
        self.doga_logs.entry_widget.scroll_exit = True
        self.doga_logs.values = []

        # setup a section for Alert history
        self.alert_history = self.window.add(npyscreen.BoxTitle, name="Alert History", max_width=75, relx=52, rely=7)
        self.alert_history.entry_widget.scroll_exit = True
        self.alert_history.values = []

        # update parent widget by embedding sub-widgets
        self.window.edit()



# -*- coding: utf-8 -*-

"""
Doga.parsers.payload

This module parse payload string and collect information from that.
info involves (method, host, resource path, section, http_type, useragent).
"""

import re
import sys


class PayloadParser:

    def __init__(self, log_generator):
        self.log_generator = log_generator

        self.method = '(GET|HEAD|POST|PUT|DELETE|TRACE|OPTIONS|CONNECT|PATCH)'
        self.path = '(\/.*)'
        self.http_type = '(HTTP\/1.[0-1])'

        self.req_regex = "%s\s%s\s%s" % (
            self.method, self.path, self.http_type)
        self.host_regex = "Host:\s(.*)\r"
        self.useragent_regex = "User-Agent:\s(.*)\r"

    def parse(self, data, ports):
        """ Parse request method, path, host, useragent, httptype etc.

        param: data(str) : packet payload string
        param: ports(list) : list object having source and destination ports
        """

        req_str = re.search(self.req_regex, data)
        host_str = re.search(self.host_regex, data)
        useragent_str = re.search(self.useragent_regex, data)

        try:
            method = req_str.group(1)
            path = req_str.group(2)
            http_type = req_str.group(3)
            host = host_str.group(1)
            useragent = useragent_str.group(1)
            section = path.split('?')[0]

            self.log_generator.generate(
                method, path, http_type, host, useragent, section)
        except:
            pass



# -*- coding: utf-8 -*-

"""
Doga.thread_jobs

This module works as 'TimeInterval' function and invoke callback functions
also run npyscreen.NPSApp as a separate thread using 'Job' class
"""

import threading


class Job(threading.Thread):
    def __init__(self, app):
        threading.Thread.__init__(self)

        self.app = app

    def run(self):
        self.app.run()


class Timer(threading.Thread):

    def __init__(self, interval, event, callback):
        threading.Thread.__init__(self)
        self.stopped = event

        self.interval = interval
        self.callback = callback

    def run(self):
        while not self.stopped.wait(self.interval):
            self.callback()



# -*- coding: utf-8 -*-

"""
Doga.logs.generator

This module manage Doga log generation and writing log to log file.
"""

import os
import time

from ..config.configer import value


class LogGenerator:

    def __init__(self, statistics, custom_file=None):
        self.statistics = statistics

        self.log_file_path = None
        self.log_file = None

        self.connect_log_file()

        self.is_custom_log = False

        self.custom_file = None

        if custom_file is not None:
            self.is_custom_log = True
            self.custom_file = custom_file

    def connect_log_file(self):
        """ manage instance variable for Doga log file path
        """

        home_dir = os.path.expanduser('~')

        self.log_file_path = os.path.join(home_dir, value('logfile'))
        self.log_file = os.path.abspath(self.log_file_path)

    def timestamp(self):
        """ return current timestamp(str) in special format to be used in logs
        """

        return time.strftime("%d/%b/%Y:%H:%M:%S %z")

    def write_log(self, log_string):
        """ write log strings to log file

        param: log_string(str): formatted log string
        """

        with open(self.log_file, 'a+') as f:
            f.write(log_string)
            f.close()

        if (self.is_custom_log):
            with open(self.custom_file, 'a+') as f:
                f.write(log_string)
                f.close()

    def generate(self, method, path, http_type, host, useragent, section):
        """ generate formatted log string for each request

        param: method(str) : request method type
        param: path(str) : resource path
        param: http_type(str) : HTTP type (1.1/1.0)
        param: host(str) : host requested
        param: useragent(str) : user-agent for requesting source
        param: section(str) : section for request's resource path
        """

        timestr = self.timestamp()
        log_str = "%s [%s] \"%s %s %s\" \"%s\"\n" % (
            host, timestr, method, path, http_type, useragent)

        self.statistics.queue_event(method, host, section)
        self.write_log(log_str)



# -*- coding: utf-8 -*-

import argparse
import sys

from statistics import Statistics
from interfaces.gui import DogaGUI
from thread_jobs import Job
from logs.generator import LogGenerator
from parsers.payload import PayloadParser
from parsers.packet import PacketParser
from interfaces.sockets import SocketInterface


def main():
    desc = 'HTTP log monitoring console for Humans'
    parser = argparse.ArgumentParser(description=desc)

    # argument for optional custom log file
    parser.add_argument(
        '-f', dest='file', type=str, default=None, help='use custom log file')

    # dictionary object with all supplied arguments
    args = vars(parser.parse_args())

    # Statistics class instance
    statistics = Statistics()

    # run npyscreen.NPSApp using separate thread
    app = DogaGUI(statistics)
    Job(app).start()

    # LogGenerator class instance
    # optional logfile is given
    if (args['file']):
        generator = LogGenerator(statistics, args['file'])
    # log to default logfile
    else:
        generator = LogGenerator(statistics)

    # PayloadParser class instance
    payload_parser = PayloadParser(generator)

    # PacketParser class instance
    packet_parser = PacketParser(payload_parser)

    # SocketInterface class interface
    socket_interface = SocketInterface(packet_parser)



from .interfaces.gui import DogaGUI
from .interfaces.sockets import SocketInterface
from .parsers.packet import PacketParser
from .parsers.payload import PayloadParser
from .logs.generator import LogGenerator
from statistics import Statistics
from thread_jobs import Job




# -*- coding: utf-8 -*-

"""
Doga.parsers.packet

This module parse traffic packet and filter HTTP packets among them.
further packets are parsed and info related to IP/TCP Headers is collected
"""

import socket
import struct

ETH_LENGTH = 14


class PacketParser:

    def __init__(self, payload_parser):
        self.payload_parser = payload_parser

    def parse_ip_header(self, packet_string):
        """ Parse required info from packret according IP Header structure
        Reference: http://en.wikipedia.org/wiki/IPv4#Header
        """

        iph = packet_string[ETH_LENGTH:20 + ETH_LENGTH]
        unpacked_iph = struct.unpack('!BBHHHBBH4s4s', iph)

        iph_first_byte = unpacked_iph[0]
        version = iph_first_byte >> 4
        ihl = iph_first_byte & 0xF
        iph_length = ihl * 4

        packet_protocol = unpacked_iph[6]

        source_addr = socket.inet_ntoa(unpacked_iph[8])
        dest_addr = socket.inet_ntoa(unpacked_iph[9])

        return (iph_length, packet_protocol, [source_addr, dest_addr])

    def parse_tcp_header(self, packet_string, iph_length):
        """ Parse required info from packet according TCP Header structure
        Reference: http://en.wikipedia.org/wiki/Transmission_Control_Protocol#TCP_segment_structure
        """

        shift = iph_length + ETH_LENGTH

        tcph = packet_string[shift:shift + 20]
        unpacked_tcph = struct.unpack('!HHLLBBHHH', tcph)

        source_port = unpacked_tcph[0]
        dest_port = unpacked_tcph[1]

        data_offset = unpacked_tcph[4]
        tcph_length = (data_offset >> 4) * 4

        header_size = ETH_LENGTH + iph_length + tcph_length

        data_size = len(packet_string) - header_size
        data = packet_string[header_size:]

        return (data, [source_port, dest_port])

    def verify_packet_data(self, ip, data, addr, ports):
        """ Verify the packet to filter HTTP packets

        param: data(str) : packet payload string
        param: addr(list) : source and destination connection addresses
        param: ports(list) : source and destination connection ports
        param: ip(str): local IP address of machine
        """

        is_http = True if 80 in ports and ip == addr[0] else False
        verified = is_http and (len(data) != 0)

        return verified

    def parse(self, ip, packet_string):
        """ Parse required info from packet according Ethernet Header structure
        Reference: http://en.wikipedia.org/wiki/Ethernet_frame#Structure

        param: ip(str): local IP address of machine
        param: packet_string(str): packet string from packet tuple object
        """

        eth_h = packet_string[:ETH_LENGTH]
        unpacked_eth_h = struct.unpack('!6s6sH', eth_h)

        eth_protocol = socket.ntohs(unpacked_eth_h[2])

        if (eth_protocol == 8):
            iph_len, packet_protocol, addr = self.parse_ip_header(
                packet_string)

            if (packet_protocol == 6):
                data, ports = self.parse_tcp_header(packet_string, iph_len)

                if self.verify_packet_data(ip, data, addr, ports):
                    self.payload_parser.parse(data, ports)



from setuptools import setup

setup(
    name='Doga',
    version='0.1.0',
    description='HTTP log monitoring console for Humans',
    keywords='http log monitor console doga',
    author='Pravendra Singh',
    author_email='hackpravj@gmail.com',
    url='https://github.com/pravj/Doga',
    license='MIT',
    packages=['Doga', 'Doga.config', 'Doga.logs', 'Doga.parsers', 'Doga.interfaces'],
    package_data = {
        'Doga': ['config/config.ini']
    },
    install_requires=[
        "npyscreen >= 4.4.1",
    ],
    entry_points={
        'console_scripts': ['doga = Doga.doga:main']
    }
)



# -*- coding: utf-8 -*-

"""
Doga.interfaces.sockets

This module does some Socket related tasks for Doga.
"""

import sys
import socket
import struct

from ..config.configer import value


class SocketInterface:

    def __init__(self, packet_parser):
        self.ip = self.ipv4()
        self.raw_socket = None

        self.packet_parser = packet_parser

        self.create_raw_socket()
        self.capture(self.raw_socket)

    def ipv4(self):
        """ return local IP Address(str) of system(IPV4 type)

        though this method depends on availability of 'httpbin.org'
        """

        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.connect((value('httpbin'), 80))

        try:
            ipv4 = sock.getsockname()[0]
        except ValueError:
            sys.exit()

        sock.close()

        return ipv4

    def create_raw_socket(self):
        """ Create row socket(SOCK_RAW type) to listen for traffic
        """

        try:
            self.raw_socket = socket.socket(
                socket.AF_PACKET, socket.SOCK_RAW, socket.ntohs(0x0003))
        except socket.error, msg:
            print "unable to create socket"
            print "Error Code %d : Message %s" % (msg[0], msg[1])
            sys.exit()

    def capture(self, sock):
        """ Capture packets in traffic

        param: sock(socket._socketobject): raw socket that listen for traffic
        """

        while True:
            # socket.recvfrom() method returns tuple object
            packet_tuple = sock.recvfrom(65565)
            packet_str = packet_tuple[0]

            self.packet_parser.parse(self.ip, packet_str)



# -*- coding: utf-8 -*-

"""
Doga.statistics

This module manage information about log statistics
update and keep track of usage summary periodically
"""

import time
import threading
from collections import Counter

from config.configer import value
from thread_jobs import Timer


class Statistics:

    def __init__(self):
        # list object having all requests of a timespan
        self.queue = []
        # total number of requests so far
        self.total = 0

        # all requests of a timespan, to be used for alert status
        self.alert_queue = []

        self.stop_event = threading.Event()

        # callback update_queue method using thread
        self.stats_timer = Timer(10, self.stop_event, self.update_queue)
        self.stats_timer.start()

        # callback update_alert_queue method using thread
        self.alert_timer = Timer(
            10, self.stop_event, self.update_alert_queue)
        self.alert_timer.start()

        # callback check_stats method using thread
        self.stats_scanner = Timer(1, self.stop_event, self.check_stats)
        self.stats_scanner.start()

        # log data and alert history values
        self.doga_logs = []
        self.alert_history = []

        # alert and usage status templates
        self.stats_template = "Maximum: [No recent requests], Recent: 0, Total: 0"
        self.alert_template = "Traffic status: Normal, Alert state: No"

        # alert state
        self.is_alert = False

        # timestamps representing alert start and end time
        self.alert_start = ""
        self.alert_end = ""

    def template(self, type):
        """ return a template value

        param: type(str): type of the template (alert or stats)
        """

        if (type == 'alert'):
            return self.alert_template
        elif (type == 'stats'):
            return self.stats_template

    def value(self, type):
        """ return value of private data

        param: type(str): type of data to return
        """

        if (type == 'logs'):
            return self.doga_logs
        elif (type == 'history'):
            return self.alert_history

    def queue_event(self, method, host, section):
        """ queue each request to be used in statistics

        param: method(str) : request method type
        param: host(str) : requested resource host
        param: section(str) : requested resource section
        """

        self.queue.append(host + section)
        self.alert_queue.append(host + section)

        self.doga_logs.append("%s %s%s" % (method, host, section))

    def update_statistics(self):
        """ update usage stats template having maximum hits and count of total hits
        """

        self.total += len(self.queue)

        status = "[No recent requests]"
        if (len(self.queue) > 0):
            counts = Counter(self.queue)
            frequency = counts.most_common()

            status = "[%s : %d]" % (frequency[0][0], frequency[0][1])

        self.stats_template = "Maximum : %s, Recent : %d, Total : %d" % (
            status, len(self.queue), self.total)

    def update_queue(self):
        """ update the doga statistics and queue periodically
        """

        self.update_statistics()
        self.queue = []

    def update_alert_queue(self):
        """ update the alert queue periodically
        """

        self.alert_queue = []

    def check_stats(self):
        """ check for requests and update the alert status according to that 
        """

        maximum = int(value('maximum'))

        # we are in alert state
        if (self.is_alert):
            if (len(self.alert_queue) < maximum):
                self.is_alert = False
                self.alert_end = time.strftime("%H:%M:%S")
                status_string = "Max Hits: %d, Recovered at: %s" % (
                    len(self.alert_queue), self.alert_start)
                self.alert_template = "Traffic: Normal, Alert state: No, %s" % (
                    status_string)
                self.alert_history.append(self.alert_template)
        # we are alert free now
        else:
            if (len(self.alert_queue) > maximum):
                self.is_alert = True
                self.alert_start = time.strftime("%H:%M:%S")
                status_string = "Max Hits: %d, Triggered at: %s" % (
                    len(self.alert_queue), self.alert_start)
                self.alert_template = "Traffic: High, Alert state: Yes, %s" % (
                    status_string)
                self.alert_history.append(self.alert_template)
